
#include <glib.h>

#include "cdaudio_stream.h"
#include "common.h"
#include "support.h"




gint checkCDMounted (void)
{
  FILE *fp;
  struct mntent *mnt;

	/* check if drive is mounted (from Mark Buckaway's cdplayer code) */
	if  ((fp = setmntent (MOUNTED, "r")) == NULL)
	{
	  g_print (_("\nCouldn't open %s :\n%s\n"), MOUNTED, strerror (errno));
	  return (0);
	}

	while ((mnt = getmntent (fp)) != NULL)
	{
		if (strcmp (mnt->mnt_type, "iso9660") == 0)
		{
		  if  (strcmp (cd_device, mnt->mnt_fsname) == 0)
			{
			  g_print (_("\nCDROM already mounted\n"));
			  endmntent (fp);
			  return (0);
			}
		}
	}
	endmntent (fp);
}


void cd_init (cdType *cdrom)
{
	memset (cdrom, 0, sizeof (cdType));
	cdrom->cd_fd = -1;
	cdrom->cur_index = 0;
	cdrom->cur_track = -1;
	cdrom->cur_balance = 10;
}




/* Read the table of contents from the CD.  Return a pointer to a cdinfo
 * struct containing the relevant information.
 * This is a static struct.  Returns NULL if there was an error */
static gint read_toc (cdType *cdrom)
{
	struct cdrom_tochdr hdr;
	struct cdrom_tocentry entry;
	gint i, pos;
	
  
  if (cdrom->cd_fd < 0)
		return (-1);

	if (ioctl (cdrom->cd_fd, CDROMREADTOCHDR, &hdr))
	{
		perror ("readtochdr");
		return (-1);
	}
	cdrom->thiscd.length = 0;
	cdrom->thiscd.ntracks = hdr.cdth_trk1;


	if (cdrom->thiscd.trk != NULL)
		free (cdrom->thiscd.trk);

	cdrom->thiscd.trk = malloc ((cdrom->thiscd.ntracks + 1) * sizeof (struct trackinfo));

	if (cdrom->thiscd.trk == NULL)
	{
		perror ("malloc");
		return (-1);
	}
	
  for (i = 0; i <= cdrom->thiscd.ntracks; i++)
	{
	  if (i == cdrom->thiscd.ntracks)
		  entry.cdte_track = CDROM_LEADOUT;
	  else
		  entry.cdte_track = i + 1;

	  entry.cdte_format = CDROM_MSF;

		if (ioctl (cdrom->cd_fd, CDROMREADTOCENTRY, &entry))
		{
		  perror ("tocentry read");
		  return (-1);
	  }

		cdrom->thiscd.trk[i].data = cdrom->thiscd.trk[i].avoid =
		      entry.cdte_ctrl & CDROM_DATA_TRACK ? 1 : 0;

		cdrom->thiscd.trk[i].length =
		      entry.cdte_addr.msf.minute * 60 +	entry.cdte_addr.msf.second;

		cdrom->thiscd.trk[i].start =
		      cdrom->thiscd.trk[i].length * 75 + entry.cdte_addr.msf.frame;


		cdrom->thiscd.trk[i].contd = 0;
		cdrom->thiscd.trk[i].track = i + 1;
		cdrom->thiscd.trk[i].section = 0;
	}

	/* Now compute actual track lengths. */
	pos = cdrom->thiscd.trk[0].length;

	for (i = 0; i < cdrom->thiscd.ntracks; i++)
	{
		cdrom->thiscd.trk[i].length = cdrom->thiscd.trk[i + 1].length - pos;
		pos = cdrom->thiscd.trk[i + 1].length;

		if (cdrom->thiscd.trk[i].data)
		  cdrom->thiscd.trk[i].length =
			      (cdrom->thiscd.trk[i + 1].start -	cdrom->thiscd.trk[i].start) * 2;
	}

	cdrom->thiscd.length = cdrom->thiscd.trk[cdrom->thiscd.ntracks].length;
	cdrom->fTOCRead = TRUE;

	return (0);
}



/* Return values:
 *      0       CD error.
 *      1       CD in drive.
 *      2       CD has just been inserted (TOC has been read)
 *      3       CD has just been removed */
gint cd_status (cdType *cdrom)
{
	gchar realname[MAXPATHLEN];
	static gint warned = 0;
	struct cdrom_subchnl sc;
	gint ret = 1;

	if (cdrom->cd_fd < 0)
	{
	  if ((cdrom->cd_fd = open (cd_device, O_RDONLY )) < 0)
		{
		  if (errno == EACCES)
			{
			  if (!warned)
				{
				  strcpy (realname, cd_device);
				  g_print (_("\nYou haven't permission to access the CD-ROM device\n"));
				  warned++;

			    return (0);
				}
			}
			else if (errno != ENXIO)
			{
			  perror (cd_device);
			  if (cdrom->thiscd.trk != NULL)
			    free (cdrom->thiscd.trk);
				return (0);
			}

			return (0);
		}
		cdrom->cur_cdmode = 5;
  }

	if (warned)
	{
	  warned = 0;
	  g_print (_("\nNext time this message won't appear\n"));
		return (0);
	}
	sc.cdsc_format = CDROM_MSF;

	if (ioctl (cdrom->cd_fd, CDROMSUBCHNL, &sc))
	{
	  cdrom->cur_cdmode = 5;
	  cdrom->cur_track = -1;
	  cdrom->cur_cdlen = cdrom->cur_tracklen = 1;
	  cdrom->cur_pos_abs = cdrom->cur_pos_rel = cdrom->cur_frame = 0;
	  return (0);
	}

	if (cdrom->cur_cdmode == DISC_EJECTED)
	{
	  cdrom->cur_pos_rel = cdrom->cur_pos_abs = cdrom->cur_frame = 0;

	  if (read_toc (cdrom) < 0)
		{
		  close (cdrom->cd_fd);
		  cdrom->cd_fd = -1;
		  return (0);
		}
		cdrom->cur_nsections = 0;
		cdrom->cur_ntracks = cdrom->thiscd.ntracks;
		cdrom->cur_cdlen = cdrom->thiscd.length;

		cdrom->cur_cdmode = 4;
		ret = 2;
	}

	switch (sc.cdsc_audiostatus)
	{
	  case CDROM_AUDIO_PLAY:
	    cdrom->cur_cdmode = 1;

      dopos:
			cdrom->cur_pos_abs = sc.cdsc_absaddr.msf.minute * 60 +
			sc.cdsc_absaddr.msf.second;
			cdrom->cur_frame = cdrom->cur_pos_abs * 75 + sc.cdsc_absaddr.msf.frame;

			/* Only look up the current track number if necessary. */
			if  ( cdrom->cur_track < 1 || cdrom->cur_frame <
			      cdrom->thiscd.trk[cdrom->cur_track - 1].start	|| cdrom->cur_frame >=
				    (cdrom->cur_track >= cdrom->cur_ntracks ?	(cdrom->cur_cdlen + 1) * 75 :
			      cdrom->thiscd.trk[cdrom->cur_track].start))
			{
			  cdrom->cur_track = 0;
			  while (cdrom->cur_track < cdrom->cur_ntracks && cdrom->cur_frame >=
			        cdrom->thiscd.trk[cdrom->cur_track].start)
			    cdrom->cur_track++;
			}

			if ( cdrom->cur_track >= 1 && sc.cdsc_trk >
			      cdrom->thiscd.trk[cdrom->cur_track - 1].track)
							cdrom->cur_track++;

			cdrom->cur_index = sc.cdsc_ind;


      doall:
      if  (cdrom->cur_track >= 1 &&  cdrom->cur_track <= cdrom->cur_ntracks)
			{
			  cdrom->cur_avoid = cdrom->thiscd.trk[ cdrom->cur_track - 1 ].avoid;
			  cdrom->cur_contd = cdrom->thiscd.trk[cdrom->cur_track - 1].contd;
			  cdrom->cur_pos_rel =
				      (cdrom->cur_frame -	cdrom->thiscd.trk[cdrom->cur_track - 1].start) / 75;
			  if (cdrom->cur_pos_rel < 0)
			    cdrom->cur_pos_rel = -cdrom->cur_pos_rel;
			}

			if (cdrom->cur_pos_abs < 0)
			  cdrom->cur_pos_abs = cdrom->cur_frame = 0;

			if (cdrom->cur_track < 1)
			  cdrom->cur_tracklen = cdrom->thiscd.length;
			else
			  cdrom->cur_tracklen = cdrom->thiscd.trk[cdrom->cur_track - 1].length;

    break;

	  case CDROM_AUDIO_PAUSED:
			if (cdrom->cur_cdmode == 1 || cdrom->cur_cdmode == 3)
			{
			  cdrom->cur_cdmode = 3;
			  goto dopos;
			}
			else
			  cdrom->cur_cdmode = 4;

			goto doall;
    break;

	  case CDROM_AUDIO_COMPLETED:
	    cdrom->cur_cdmode = 0;     /* waiting for next track. */
	  break;

	  case CDROM_AUDIO_NO_STATUS:
	    cdrom->cur_cdmode = 4;
			cdrom->cur_lasttrack = cdrom->cur_firsttrack = -1;
			goto doall;
	  break;
  }

	return (ret);
}



/* Pause the CD, if it's in play mode. If it's already paused, go back to
 * play mode */
gint cd_pause (cdType *cdrom)
{
  if (cdrom->cd_fd < 0)
    return (DISC_ERROR);

  switch (cdrom->cur_cdmode)
	{
	  case DISC_PLAYING:
	    cdrom->cur_cdmode = 3;
			ioctl (cdrom->cd_fd, CDROMPAUSE);
	  break;

	  case DISC_PAUSED:
	    cdrom->cur_cdmode = 1;
			ioctl (cdrom->cd_fd, CDROMRESUME);
	  break;
	}
}



/* Stop the CD if it's not already stopped */
gint cd_stop (cdType *cdrom)
{
  if (cdrom->cd_fd < 0)
    return (DISC_ERROR);

  if (cdrom->cur_cdmode != 4)
	{
	  cdrom->cur_lasttrack = cdrom->cur_firsttrack = -1;
	  cdrom->cur_cdmode = 4;
	  ioctl (cdrom->cd_fd, CDROMSTOP);
	  cdrom->cur_track = 1;
	}
}




/* Play the CD from one position to another (both in frames) */
static gint play_chunk (cdType *cdrom, gint start, gint end)
{
  struct cdrom_msf msf;

  if  (!cdrom->fTOCRead  ||  cdrom->cd_fd < 0)
    return (DISC_ERROR);

	end--;
	if (start >= end)
		start = end - 1;

	msf.cdmsf_min0 = start / (60 * 75);
	msf.cdmsf_sec0 = (start % (60 * 75)) / 75;
	msf.cdmsf_frame0 = start % 75;
	msf.cdmsf_min1 = end / (60 * 75);
	msf.cdmsf_sec1 = (end % (60 * 75)) / 75;
	msf.cdmsf_frame1 = end % 75;

	if (ioctl (cdrom->cd_fd, CDROMSTART))
	{
		perror ("CDROMSTART");
		return;
	}

	if (ioctl (cdrom->cd_fd, CDROMPLAYMSF, &msf))
	{
		perror ("CDROMPLAYMSF");
		return;
	}
}




/* Start playing the CD or jump to a new position.
 * "pos" is in seconds, relative to start of track */
gint cd_play (cdType *cdrom, gint start, gint pos, gint end)
{
  if (!cdrom->fTOCRead || cdrom->cd_fd < 0)
    return (DISC_ERROR);

  cdrom->cur_firsttrack = start;
  start--;
  end--;
  cdrom->cur_lasttrack = end;

  play_chunk (cdrom, cdrom->thiscd.trk[start].start + pos * 75,
        end >= cdrom->cur_ntracks ?	cdrom->cur_cdlen * 75 :
				cdrom->thiscd.trk[end].start - 1);
}
